/*
 * Copyright (c) 2001 VM Labs, Inc.  All rights reserved.
 *
 * NOTICE: VM Labs permits you to use, modify, and distribute this file
 * in accordance with the terms of the VM Labs license agreement
 * accompanying it. If you have received this file from a source other
 * than VM Labs, then your use, modification, or distribution of it
 * requires the prior written permission of VM Labs.
*/

//==================================================================================================
// Please widen your screen to fit 100 characters per line and set tab stops to 2 (if necessary)
//==================================================================================================
// 4567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//==================================================================================================

//==================================================================================================
// Includes
//==================================================================================================

#include <stdio.h>
#include <stdlib.h>
#include <nuon/sprite.h>
#include <nuon/termemu.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/video.h>
#include <nuon/joystick.h>
#include <nuon/nise.h>
#include <nuon/mutil.h>
#include <nuon/msprintf.h>

/*
    Level definitions
*/

#include "levels.h"
#include "debug.h"
#include "custom.h"

// Bunch of constants

#define SCREEN_WIDTH  360
#define SCREEN_HEIGHT 240
#define DMAFLAGS  (DMA_PIXEL_WRITE | ((SCREEN_WIDTH/8)<<16) | (4<<4) | (1<<11))

#define DEADZONE_DIAGONAL 16
#define DEADZONE_CENTER   32

#define TILESIZE 8
#define INC 1
#define INCG 1

#define TRUE  1
#define FALSE 0

#define OFFX 37
#define OFFY 20

#define SPRITESIZE  32
#define CHOMPERSIZE 40
#define ANIM_DELAY  1

#define GHOSTSCALE   0x7000
#define CHOMPERSCALE 0x7000
#define ITEMSCALE    0x5000
#define SCORESCALE   0x8000
#define LIVESSCALE   0x6000

#define CHOMPER_ANIM_FRAMES 4

#define PILL_OFFSET_Y       0
#define POWERPILL_OFFSET_Y  SPRITESIZE
#define BLANK_OFFSET_X      (4 * SPRITESIZE)
#define BLANK_OFFSET_Y      (0 * SPRITESIZE)
#define PROGRESS_OFFSET_X   (4 * SPRITESIZE)

#define SCOREX 40
#define SCOREY 215

#define CHASEBARX 180
#define CHASEBARY 215

#define DELAY 15

#define RATIO 2


#define NUM_SPRITES 16
#define NUM_BUFFERS 3
#define MAX_GHOSTS  6

//==================================================================================================
// Binaries (see data.s)
//==================================================================================================
extern char playerImages[];

extern char miscImages[];
extern char StartScreen[];
extern char GameOverScreen[];
extern char GameCompleteScreen[];
extern char LevelCompleteScreen[];
extern char GameBeginWav[];
extern char ChomperChompWav[];
extern char FruitEatWav[];
extern char ChaseWav[];
extern char GhostEatWav[];
extern char ChomperDiesWav[];
extern char ExtraChomperWav[];
extern char Digits[];
extern char PauseImage[];

//==================================================================================================
// Structures for the ghosts and chomperman
//==================================================================================================

typedef struct
{
  SPRITE *ghost;
  int gphase;
  int startx, starty;
  int ghostx, ghosty;
  int nghostx, nghosty;
  int ghostAx, ghostAy;
  int gincx, gincy;
  int way;
  int last_way;
  int hist[16], in, out;
  int active;
} GHOST;


typedef struct
{
  SPRITE *chomperman;
  int mphase, aphase;
  int animDelay;
  int startx, starty;
  int chomperx, chompery;
  int nchomperx, nchompery;
  int chomperAx, chomperAy;
  int incx, incy;
  int angle;
  int direction, ndir;
} CHOMPERMAN;

//==================================================================================================
// Local prototypes
//==================================================================================================

int PlayWAV(void *wavaddr, int voice, PCMPOS *Pan, int volume, int echo);
void InitGhost(int i);
void WaitForJoystick(int but);
void PauseGame (void);	//SRA 1/14/2002 - pauses game, displays animated logo


//==================================================================================================
// Globals
//==================================================================================================

// SDC 2002-02-10:  Broke up large sprite image into smaller pieces for customization
sCustomImg aPlayerImg[1] = {
  { playerImages },
};
sCustomImg miscImg;
sCustomImg digitsImg = { Digits };

SPR_IMAGE_INFO StartScr;
SPR_IMAGE_INFO GameOverScr;
SPR_IMAGE_INFO CompleteScr;
SPR_IMAGE_INFO LevelScr;
SPR_IMAGE_INFO PauseScr;
PCMPOS Pan;

void *FrameBuffer[NUM_BUFFERS],*MazeBuf;
static int fps;
int drawBuffer = 0;
char *localScratch;
int localSize;
SPRITE *bonusSprite, *PauseSprite;
int saphase;

SPRITE *sprProgress;
int progressRot;

GHOST ghosts[MAX_GHOSTS];
CHOMPERMAN chomper;

int score;
int extra;
int lives;
int gameover;
int chase;
int pills;
int collected;
int delay;
int level;
int ghostsEaten;
int gDivider,gDividerCnt;
int pDivider,pDividerCnt;
int numGhosts;
int chasevoice;

int wallMax, pillMax, fruitMax, ghostMax;
int wallIndex, pillIndex, fruitIndex, ghostIndex;
int debug;

unsigned char maze[MAZE_W*MAZE_H];

//
// For DEBUG
//
int retVal;

/*
    Copy via DMA
*/


void CopyMaze(char *d,char *s,int pages)
{
  while(pages)
    {
      _DMALinear((64<<16)+(1<<13),s,localScratch);
      _DMALinear((64<<16),d,localScratch);
      pages--;
      s=&s[256];
      d=&d[256];
    }
}

/*
    Simple routine to draw a filled rectangle (via DMA)
*/

void myFillRect(char *screen, int dmaflags, int initx, int inity, int wide, int high, int color)
{
  int x, y;
  int lastx, lasty;
 
  lastx = initx + wide;
  lasty = inity + high;
  
  for (x = initx; x < lastx; x += 8)
    for (y = inity; y < lasty; y += 8)
      _raw_plotpixel(dmaflags, (void *) screen, (8<<16) | x, (8<<16) | y, color);
}

/*
    Convert ASCII description of a maze inta a real framebuffer
*/

// SDC 2002-02-10:  Broke up sprite image file into separate images for further customization
//                  Rewrote '#' case to fit the new sprite layout
// SDC 2002-02-14:  Rewrote so that every tile gets drawn with "blank" tile first

void CreateMaze(SPR_IMAGE_INFO *imgWalls, SPR_IMAGE_INFO *imgPills, SPR_IMAGE_INFO *imgFruit,
		long *frame, char *maze, int w, int h)
{
  char mazeChar;
  int x, y;
  int wallOffsetX, wallOffsetY;

  pills = numGhosts = 0;
  chomper.startx = 18;      
  chomper.starty = 19;      
  
  for (x = 0; x < w; x++)
    for (y = 0; y < h; y++) {

      mazeChar = maze[x + (y * w)];

      // Clear the tile first with the blank tile
      SPRBlitter(imgWalls->img, imgWalls->dmaflags, BLANK_OFFSET_X, BLANK_OFFSET_Y,
		 SPRITESIZE, SPRITESIZE,
		 (char *) frame, DMAFLAGS, (x * TILESIZE) + OFFX, (y * TILESIZE) + OFFY,
		 (TILESIZE * 0x10000) / SPRITESIZE, (TILESIZE * 0x10000) / SPRITESIZE, 0,
		 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);

      switch(mazeChar) {
        case '#':
	  wallOffsetX = wallOffsetY = 0;
	  if ((y != (h-1)) && (maze[ x    + ((y+1)*w)] == '#')) wallOffsetY += 1;
	  if ((x != (w-1)) && (maze[(x+1) + ( y   *w)] == '#')) wallOffsetY += 2;
	  if ( y           && (maze[ x    + ((y-1)*w)] == '#')) wallOffsetX += 1;
	  if ( x           && (maze[(x-1) + ( y   *w)] == '#')) wallOffsetX += 2;
	  SPRBlitter(imgWalls->img, imgWalls->dmaflags, wallOffsetX * SPRITESIZE,
		     wallOffsetY * SPRITESIZE, SPRITESIZE, SPRITESIZE,
		     (char *) frame, DMAFLAGS, (x * TILESIZE) + OFFX, (y * TILESIZE) + OFFY,
		     (TILESIZE * 0x10000) / SPRITESIZE, (TILESIZE * 0x10000) / SPRITESIZE, 0,
		     0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
	  break;

        case ' ':
	  break;

        case '.':
	  pills++;
	  SPRBlitter(imgPills->img, imgPills->dmaflags, 0, PILL_OFFSET_Y,
		     SPRITESIZE, SPRITESIZE,
		     (char *) frame, DMAFLAGS, (x * TILESIZE) + OFFX, (y * TILESIZE) + OFFY,
		     (TILESIZE * 0x10000) / SPRITESIZE, (TILESIZE * 0x10000) / SPRITESIZE, 
		     0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
	  break;

        case '*':
	  pills++;
	  SPRBlitter(imgFruit->img, imgFruit->dmaflags, 0, 0, SPRITESIZE, SPRITESIZE,
		     (char *) frame, DMAFLAGS, (x * TILESIZE) + OFFX, (y * TILESIZE) + OFFY,
		     (TILESIZE * 0x10000) / SPRITESIZE, (TILESIZE * 0x10000) / SPRITESIZE, 
		     0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
	  break;

        case 'o':
	  pills++;
	  SPRBlitter(imgPills->img, imgPills->dmaflags, 0, POWERPILL_OFFSET_Y,SPRITESIZE,SPRITESIZE,
		     (char *) frame, DMAFLAGS, (x * TILESIZE) + OFFX, (y * TILESIZE) + OFFY,
		     (TILESIZE * 0x10000) / SPRITESIZE, (TILESIZE * 0x10000) / SPRITESIZE, 
		     0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
	  break;

        case 'X':
	  chomper.startx = x;
	  chomper.starty = y;
	  break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
	  ghosts[mazeChar - '1'].startx = x;
	  ghosts[mazeChar - '1'].starty = y;
	  if ((mazeChar - '1' + 1) > numGhosts) numGhosts = mazeChar - '1' + 1;
	  break;
      }       
    }
}

/*
  Erase a "pill" from maze and screen
*/

// SDC 2002-02-14:  Changed from erasing the tile to copying the blank tile over

void DeletePill(SPR_IMAGE_INFO *img, long *frame, int x, int y)
{
  SPRBlitter(img->img, img->dmaflags, BLANK_OFFSET_X, BLANK_OFFSET_Y, SPRITESIZE, SPRITESIZE,
	     (char *) frame, DMAFLAGS, (x * TILESIZE) + OFFX, (y * TILESIZE) + OFFY,
	     (TILESIZE * 0x10000) / SPRITESIZE, (TILESIZE * 0x10000) / SPRITESIZE, 0,
	     0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
  maze[x + (y * MAZE_W)] = ' ';
}

/*
 * wait for video
 * returns the number of elapsed fields
 * since the last sync
 */

int VideoSync(void)
{
  static int frames;
  static int lastfield;
  static int lastsecond;
  int ret;

  int curfield;

  // we want to view the buffer we just finished drawing

  // set the video to point at the drawn buffer
  _VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[drawBuffer]);
  
  // find out how many fields have been drawn
  curfield = _VidSync(-1);

  // remember that we've drawn a frame
  frames++;

  // have more than 60 fields (1 second) passed?
  if (curfield - lastsecond >= 60) {
    fps = frames;
    frames = 0;
    lastsecond = curfield;
  }

  // draw to a new buffer
  drawBuffer++;
  if (drawBuffer >= NUM_BUFFERS)
    drawBuffer = 0;

#if NUM_BUFFERS == 2
  // wait until a new field starts (so it's no longer showing the field we are going to draw on)
  curfield = _VidSync(0);
#endif

  ret = curfield - lastfield;
  while (ret == 0) {
    curfield = _VidSync(-1);
    ret = curfield - lastfield;
  }
  lastfield = curfield;
  return ret;
}

//==================================================================================================
// getFirstTGAPixel
//
// Parses a TGA file to find the first pixel color within the image.  Returns 0 (black) on failure,
// or the color upon success.
//
// Revision history
//   SDC 2002-02-14:  Initial revision
//==================================================================================================

int getFirstTGAPixel(char *tgaFile)
{
  int offsetColorMap;
  int offsetPixels;
  int colorMapStart;
  int colorMapLength;
  int colorBytes;
  unsigned char index;
  unsigned char r, g, b;

  offsetColorMap = offsetPixels = 18 + tgaFile[0];

  if (tgaFile[1] != 1) return(0);  // Must have color map
  if (tgaFile[2] != 9) return(0);  // Must be RLE color-mapped

  colorMapStart  = tgaFile[3] + (tgaFile[4] << 8);
  colorMapLength = tgaFile[5] + (tgaFile[6] << 8);

  if ((tgaFile[7] != 24) && (tgaFile[7] != 32)) return(0);  // Colors must be 24-bit or 32-bit
  if (tgaFile[16] != 8) return(0);  // Pixels must be 8-bit color depth

  colorBytes = (tgaFile[7] == 24) ? 3 : 4;
  offsetPixels += colorMapLength * colorBytes;

  index = tgaFile[offsetPixels+1];  // Skip past RLE opcode to get first color index
  r = tgaFile[offsetColorMap + (index*colorBytes) + 2];
  g = tgaFile[offsetColorMap + (index*colorBytes) + 1];
  b = tgaFile[offsetColorMap + (index*colorBytes) + 0];

  return((((int) r) << 24) | (((int) g) << 16) | (((int) b) << 8));
}


/*
    Start a score animation
*/

void StartScoreAnimation(int s,int x, int y)
{
  if (bonusSprite)
    SPRSetSpriteSource(bonusSprite, s * SPRITESIZE, 0, SPRITESIZE, SPRITESIZE);
  else
    bonusSprite = SPRCreateSprite(&miscImg.spr, s * SPRITESIZE, 0, SPRITESIZE, SPRITESIZE);
  
  SPRAddSprite(bonusSprite, x * TILESIZE + OFFX, y * TILESIZE + OFFY, 0, SCORESCALE, SCORESCALE,
	       0x40000002, kBlack, 1);
  SPRSetSpriteXY(bonusSprite,x * TILESIZE + OFFX, y * TILESIZE + OFFY);
  
  saphase=60;             
}

/*
    Continue score animation
*/

void DoScoreAnimation(void)
{
  if (saphase) {
    SPRSetSpriteScale(bonusSprite, SCORESCALE + (60 - saphase) * 0x444,
		      SCORESCALE + (60 - saphase) * 0x444);
    SPRSetSpriteRotation(bonusSprite, (60 - saphase) * 0x800);
    saphase--;
    if (!saphase) SPRRemoveSprite(bonusSprite); 
  }
}

/*
    Start dead ghost animation
*/

void StartGhostDeadAnimation(int ghostNum, int x, int y)
{
  ghosts[ghostNum].active = 0;
    
  ghosts[ghostNum].ghostAx = (x * TILESIZE) + OFFX;
  ghosts[ghostNum].ghostAy = (y * TILESIZE) + OFFY;
  ghosts[ghostNum].gphase  = 60;
  ghosts[ghostNum].gincx   = ((ghosts[ghostNum].startx-x)*TILESIZE) / ghosts[ghostNum].gphase;
  ghosts[ghostNum].gincy   = ((ghosts[ghostNum].starty-y)*TILESIZE) / ghosts[ghostNum].gphase;
  SPRSetSpriteSource(ghosts[ghostNum].ghost, 6 * SPRITESIZE, 0, SPRITESIZE, SPRITESIZE);
  SPRSetSpriteScale(ghosts[ghostNum].ghost, 0x20000 / RATIO, 0x20000 / RATIO);
}

/*
    Control Chomperman (by joystick or remote control)
*/

// SDC 2002-02-10:  Changed control so no diagonals are allowed
//                  Made chomper sprite rotate so can reuse same animation for all 4 directions
// SDC 2002-02-15:  Added support for analog stick

void ControlChomperman(void)
{
  int edge,joystick,i;
  static int old_joystick;
  char xAxis, yAxis;
        
  // logical OR between joystick and remote
  joystick = _Controller[1].buttons | _Controller[0].buttons;

  xAxis = _Controller[1].d1.xAxis;
  yAxis = _Controller[1].d2.yAxis;
  // If joystick is too close to diagonal directions then skip.  Prevents jitter from causing
  // direction to change unexpectedly.
  //
  // The diagonals occur when:  abs(xAxis) == abs(yAxis)
  //
  // To add some cushion, compute absolute difference between these two and ensure it is more
  // than a given deadzone.
  if (abs(abs(xAxis) - abs(yAxis)) > DEADZONE_DIAGONAL) {
    if (xAxis < -DEADZONE_CENTER) {
      if (yAxis < 0) joystick |= (-xAxis > -yAxis) ? CTRLR_DPAD_LEFT  : CTRLR_DPAD_DOWN;
      else           joystick |= (-xAxis >  yAxis) ? CTRLR_DPAD_LEFT  : CTRLR_DPAD_UP;

    } else if (xAxis > DEADZONE_CENTER) {
      if (yAxis < 0) joystick |= ( xAxis > -yAxis) ? CTRLR_DPAD_RIGHT : CTRLR_DPAD_DOWN;
      else           joystick |= ( xAxis >  yAxis) ? CTRLR_DPAD_RIGHT : CTRLR_DPAD_UP;

    } else if (yAxis < -DEADZONE_CENTER) {
      // Take advantage of the fact that previous "ifs" did not trigger (x-axis is within deadzone)
      joystick |= CTRLR_DPAD_DOWN;

    } else if (yAxis >  DEADZONE_CENTER) {
      // Take advantage of the fact that previous "ifs" did not trigger (x-axis is within deadzone)
      joystick |= CTRLR_DPAD_UP;
    }
  }

  // Clear out all but the direction bits
  joystick &= CTRLR_DPAD_UP | CTRLR_DPAD_DOWN | CTRLR_DPAD_LEFT | CTRLR_DPAD_RIGHT;

  edge = joystick ^ old_joystick;

  if (edge) chomper.ndir = joystick;

  if (chomper.mphase) {
    chomper.chomperAx += chomper.incx;
    chomper.chomperAy += chomper.incy;
    SPRSetSpriteXY(chomper.chomperman, chomper.chomperAx, chomper.chomperAy);

    // Rotate/scale chomper sprite depending on which direction it's facing
    if (chomper.angle == 0x0000) {
      // Chomper is facing right - no rotation or mirroring
      SPRSetSpriteRotation(chomper.chomperman, 0);
      SPRSetSpriteScale(chomper.chomperman, CHOMPERSCALE, CHOMPERSCALE);

    } else if (chomper.angle == 0x8000) {
      // Chomper is facing left - no rotation, but mirror x-direction
      SPRSetSpriteRotation(chomper.chomperman, 0);
      SPRSetSpriteScale(chomper.chomperman, -CHOMPERSCALE, CHOMPERSCALE);

    } else {
      // Chomper is facing up or down - rotate, but no mirroring
      SPRSetSpriteRotation(chomper.chomperman, chomper.angle);
      SPRSetSpriteScale(chomper.chomperman, CHOMPERSCALE, CHOMPERSCALE);
    }

    chomper.mphase--;
    if (!chomper.mphase) {
      chomper.chomperx = chomper.nchomperx;
      chomper.chompery = chomper.nchompery;
      chomper.incx = chomper.incy = 0;

      for(i = 0; i < numGhosts; i++)
	if (ghosts[i].active &&
	    (ghosts[i].ghostx == chomper.chomperx) &&
	    (ghosts[i].ghosty == chomper.chompery)) {
	  if (chase) {
	    ghostsEaten++;
	    PlayWAV((void *) GhostEatWav, -1, &Pan, 0x40000000, 0x40000000);
	    score += (ghostsEaten > 2) ? 800 : 400;
	    StartScoreAnimation((ghostsEaten > 2) ? 3 : 2, chomper.chomperx, chomper.chompery);
	    StartGhostDeadAnimation(i, chomper.chomperx, chomper.chompery);
	  } else {
	    PlayWAV((void *) ChomperDiesWav, -1, &Pan, 0x40000000, 0x40000000);
	    gameover = 1;
	  }
	}


      if (maze[chomper.chomperx + MAZE_W * chomper.chompery] == '.') {
	PlayWAV((void *) ChomperChompWav, -1, &Pan, 0x40000000, 0x40000000);
	DeletePill(&aWallImg[wallIndex].spr, MazeBuf, chomper.chomperx, chomper.chompery);
	score += 10;
	collected++;
      }

      if (maze[chomper.chomperx + MAZE_W * chomper.chompery]=='*') {
	PlayWAV((void *) FruitEatWav, -1, &Pan, 0x40000000, 0x40000000);
	DeletePill(&aWallImg[wallIndex].spr, MazeBuf, chomper.chomperx, chomper.chompery);
	score += 100;
	StartScoreAnimation(0, chomper.chomperx, chomper.chompery);
	collected++;
      }

      if ((maze[chomper.chomperx + MAZE_W * chomper.chompery] == 'o') && !gameover) {
	DeletePill(&aWallImg[wallIndex].spr, MazeBuf, chomper.chomperx, chomper.chompery);
	score += 200;
	StartScoreAnimation(1, chomper.chomperx, chomper.chompery);
	if (!chase) {
	  chasevoice = PlayWAV((void *) ChaseWav, 1, &Pan, 0x40000000, 0x40000000);
	  ghostsEaten = 0;
	}
	chase = 5 * 60;
	collected++;
	// Prevent a crash from eating a power pill if already in chase mode
      }
    }

  } else {
    // Check if Chomperman changes direction
    if (chomper.ndir == CTRLR_DPAD_RIGHT) {
      if (maze[(chomper.chomperx + 1) + MAZE_W * chomper.chompery] != '#')
	chomper.direction = CTRLR_DPAD_RIGHT;
	  
    } else if (chomper.ndir == CTRLR_DPAD_LEFT) {
      if (maze[(chomper.chomperx - 1) + MAZE_W * chomper.chompery] != '#')
	chomper.direction = CTRLR_DPAD_LEFT;
	  
    } else if (chomper.ndir == CTRLR_DPAD_DOWN) {
      if (maze[chomper.chomperx + MAZE_W * (chomper.chompery + 1)] != '#')
	chomper.direction = CTRLR_DPAD_DOWN;
	  
    } else if (chomper.ndir == CTRLR_DPAD_UP) {
      if (maze[chomper.chomperx + MAZE_W * (chomper.chompery - 1)] != '#')
	chomper.direction = CTRLR_DPAD_UP;
    }

    // Move Chomperman
    if (chomper.direction & CTRLR_DPAD_RIGHT) {
      if (maze[(chomper.chomperx + 1) + MAZE_W * chomper.chompery] != '#') {
	chomper.nchomperx = chomper.chomperx + 1;
	chomper.incx = INC;
	chomper.angle = 0;
	chomper.mphase = TILESIZE / INC;
      } else {
	chomper.direction = 0;
      }

    } else if (chomper.direction & CTRLR_DPAD_LEFT) {
      if (maze[(chomper.chomperx - 1) + MAZE_W * chomper.chompery] != '#') {
	chomper.nchomperx = chomper.chomperx - 1;
	chomper.incx = -INC;
	chomper.angle = 0x8000;
	chomper.mphase = TILESIZE / INC;
      } else {
	chomper.direction = 0;
      }

    } else if (chomper.direction & CTRLR_DPAD_DOWN) {
      if (maze[chomper.chomperx+MAZE_W * (chomper.chompery + 1)] != '#') {
	chomper.nchompery = chomper.chompery + 1;
	chomper.incy = INC;
	chomper.angle = 0x4000;
	chomper.mphase = TILESIZE / INC;
      }	else {
	chomper.direction = 0;
      }

    } else {
      if (chomper.direction & CTRLR_DPAD_UP) {
	if (maze[chomper.chomperx + MAZE_W * (chomper.chompery - 1)] != '#') {
	  chomper.nchompery = chomper.chompery - 1;
	  chomper.incy = -INC;
	  chomper.angle = 0xc000;
	  chomper.mphase = TILESIZE/INC;
	} else {
	  chomper.direction = 0;
	}
      }
    }
    chomper.chomperAx = chomper.chomperx * TILESIZE + OFFX;
    chomper.chomperAy = chomper.chompery * TILESIZE + OFFY;
  }

  old_joystick = joystick;
}

/*
    Ghost AI
*/

void ControlGhost(int ghostNum)
{
  int x, y, c, visual, i;
  int ways, pw[4], sight[4];
  char savemaze[1024];            
  int destx, desty;
  
  if (ghosts[ghostNum].gphase) {
      ghosts[ghostNum].ghostAx += ghosts[ghostNum].gincx;
      ghosts[ghostNum].ghostAy += ghosts[ghostNum].gincy;
      SPRSetSpriteXY(ghosts[ghostNum].ghost, ghosts[ghostNum].ghostAx,
		     ghosts[ghostNum].ghostAy);
      ghosts[ghostNum].gphase--;
      if (!ghosts[ghostNum].gphase) {
	if (!ghosts[ghostNum].active) InitGhost(ghostNum);
	  
	ghosts[ghostNum].ghostx = ghosts[ghostNum].nghostx;
	ghosts[ghostNum].ghosty = ghosts[ghostNum].nghosty;
	ghosts[ghostNum].gincx  = ghosts[ghostNum].gincy = 0;
	  
	if (ghosts[ghostNum].active)
	  if ((ghosts[ghostNum].ghostx == chomper.chomperx) &&
	      (ghosts[ghostNum].ghosty == chomper.chompery)) {
	    if (chase) {
	      ghostsEaten++;
	      PlayWAV((void *) GhostEatWav, -1, &Pan, 0x40000000, 0x40000000);
	      score += (ghostsEaten > 2) ? 800 : 400;
	      StartScoreAnimation((ghostsEaten > 2) ? 3 : 2, chomper.chomperx, chomper.chompery);
	      StartGhostDeadAnimation(ghostNum, chomper.chomperx, chomper.chompery);
	    } else {
	      PlayWAV((void *) ChomperDiesWav, -1, &Pan, 0x40000000, 0x40000000);
	      gameover = 1;
	    }
	  }
      }

  } else {
      
      destx = chomper.chomperx;
      desty = chomper.chompery;
      
      if (chase) {
	destx = MAZE_W - destx;
	desty = MAZE_H - destx;
      }
      
      x = ghosts[ghostNum].ghostx;
      y = ghosts[ghostNum].ghosty;
      
      /* Do we have visual contact with chomperman?*/
      visual = 0;

      // Check above
      x = ghosts[ghostNum].ghostx;     
      y = ghosts[ghostNum].ghosty;     
      sight[3] = 0;

      do {
	sight[3]++;
	y--;
	c = maze[x + (y * MAZE_W)];
	if (c != '#') {
	  if ((y == desty) && (x == destx)) {
	    visual = 1;
	    ghosts[ghostNum].way = 3;
	    break;
	  }
	}
      } while (c != '#');

      // Check below
      if (!visual) {
	x = ghosts[ghostNum].ghostx;     
	y = ghosts[ghostNum].ghosty;     
	sight[2] = 0;
	do {
	  sight[2]++;
	  y++;
	  c = maze[x + (y * MAZE_W)];
	  if (c != '#')	{
	    if ((y == desty) && (x == destx)) {
	      visual = 1;
	      ghosts[ghostNum].way = 2;
	      break;
	    }
	  }
	} while (c!='#');
      }

      // Check to left
      if (!visual) {
	x = ghosts[ghostNum].ghostx;     
	y = ghosts[ghostNum].ghosty;     
	sight[1] = 0;
	do {
	  sight[1]++;
	  x--;
	  c = maze[x + (y * MAZE_W)];
	  if (c != '#') {
	    if ((y == desty) && (x == destx)) {
	      visual = 1;
	      ghosts[ghostNum].way = 1;
	      break;
	    }
	  }
	} while (c != '#');
      }

      // Check to right
      if (!visual) {
	x = ghosts[ghostNum].ghostx;     
	y = ghosts[ghostNum].ghosty;     
	sight[0] = 0;
	do {
	  sight[0]++;
	  x++;
	  c = maze[x + (y * MAZE_W)];
	  if (c != '#') {
	    if ((y == desty) && (x == destx)) {
	      visual = 1;
	      ghosts[ghostNum].way = 0;
	      break;
	    }
	  }
	} while (c!='#');
      }
      
      // Go into direction of ChomperMan if possible

      if (!visual) {
	i = (rand() % 3) + 1;
	x = ghosts[ghostNum].ghostx;     
	y = ghosts[ghostNum].ghosty;     
	if ((sight[0] > i) && (ghosts[ghostNum].last_way != 1) && (destx > x)) {
	  visual = 1;
	  ghosts[ghostNum].way = 0;
	}
	if ((sight[1] > i) && (ghosts[ghostNum].last_way != 0) && (destx < x)) {
	  visual = 1;
	  ghosts[ghostNum].way = 1;
	}
	if ((sight[2] > i) && (ghosts[ghostNum].last_way != 3) && (desty > y)) {
	  visual = 1;
	  ghosts[ghostNum].way = 2;
	}
	if ((sight[3] > i) && (ghosts[ghostNum].last_way != 2) && (desty < y)) {
	  visual = 1;
	  ghosts[ghostNum].way = 3;
	}
      }

      // Go somewhere where we never have been (the last 16 moves)

      memcpy(savemaze, maze, MAZE_W * MAZE_H);
      if (!visual) {
	i = ghosts[ghostNum].out;
	while(i != ghosts[ghostNum].in) {
	  maze[ghosts[ghostNum].hist[i]] = '#';
	  i++;
	  if (i == 16) i = 0;
	}
	  
	/* Now pick random direction*/
	  
	ways = 0;
	if (maze[(x+1) + ( y    * MAZE_W)] != '#') pw[ways++] = 0;
	if (maze[(x-1) + ( y    * MAZE_W)] != '#') pw[ways++] = 1;
	if (maze[ x    + ((y+1) * MAZE_W)] != '#') pw[ways++] = 2;
	if (maze[ x    + ((y-1) * MAZE_W)] != '#') pw[ways++] = 3;
	  
	memcpy(maze, savemaze, MAZE_W*MAZE_H);
	if (!ways) {
	  if (maze[(x+1) + ( y    * MAZE_W)] != '#') pw[ways++] = 0;
	  if (maze[(x-1) + ( y    * MAZE_W)] != '#') pw[ways++] = 1;
	  if (maze[ x    + ((y+1) * MAZE_W)] != '#') pw[ways++] = 2;
	  if (maze[ x    + ((y-1) * MAZE_W)] != '#') pw[ways++] = 3;
	}
	  
	ghosts[ghostNum].way = pw[rand() % ways];
      }       
      
                        
      x = ghosts[ghostNum].ghostx;     
      y = ghosts[ghostNum].ghosty;

      
      switch(ghosts[ghostNum].way) {
	  
        case 0:
	  if (maze[(x+1)+y*MAZE_W] != '#') {
	    ghosts[ghostNum].gphase  = TILESIZE / INCG;
	    ghosts[ghostNum].gincx   = INCG;
	    ghosts[ghostNum].nghostx = x+1;
	  } else
	    ghosts[ghostNum].way     = -1;
	  break;

	case 1:
	  if (maze[(x-1)+y*MAZE_W] != '#') {
	    ghosts[ghostNum].gphase  = TILESIZE / INCG;
	    ghosts[ghostNum].gincx   = -INCG;
	    ghosts[ghostNum].nghostx = x-1;
	  } else
	    ghosts[ghostNum].way     = -1;
	  break;

	case 2:
	  if (maze[x+(y+1)*MAZE_W] != '#') {
	    ghosts[ghostNum].gphase  = TILESIZE / INCG;
	    ghosts[ghostNum].gincy   = INCG;
	    ghosts[ghostNum].nghosty = y+1;
	  } else
	    ghosts[ghostNum].way     = -1;
	  break;

	case 3:
	  if (maze[x+(y-1)*MAZE_W] != '#') {
	    ghosts[ghostNum].gphase  = TILESIZE / INCG;
	    ghosts[ghostNum].gincy   = -INCG;
	    ghosts[ghostNum].nghosty = y-1;
	  } else
	    ghosts[ghostNum].way     = -1;
	  break;
      }

      ghosts[ghostNum].last_way = ghosts[ghostNum].way;               
      
      
      i = ghosts[ghostNum].in;
      ghosts[ghostNum].hist[i] = ghosts[ghostNum].nghostx + (MAZE_W*ghosts[ghostNum].nghosty);
      i++;
      if (i == 16) i = 0;
      if (i == ghosts[ghostNum].out) {
	ghosts[ghostNum].out++;
	if (ghosts[ghostNum].out == 16) ghosts[ghostNum].out = 0;
      }
      ghosts[ghostNum].in = i;
    }
}

/*
  Initialize ghosts
*/


void InitGhost(int ghostNum)
{
  ghosts[ghostNum].active  = 1;
  ghosts[ghostNum].way     = -1;

  ghosts[ghostNum].gphase  = ghosts[ghostNum].gincx  = ghosts[ghostNum].gincy = 0;

  ghosts[ghostNum].nghosty = ghosts[ghostNum].ghosty = ghosts[ghostNum].starty;
  ghosts[ghostNum].nghostx = ghosts[ghostNum].ghostx = ghosts[ghostNum].startx;

  ghosts[ghostNum].ghostAy = ghosts[ghostNum].ghosty * TILESIZE + OFFY;
  ghosts[ghostNum].ghostAx = ghosts[ghostNum].ghostx * TILESIZE + OFFX;

  ghosts[ghostNum].hist[0] = ghosts[ghostNum].ghostx + (MAZE_W * ghosts[ghostNum].ghosty);

  ghosts[ghostNum].in = ghosts[ghostNum].out = 0;
  if (SPRAddSprite(ghosts[ghostNum].ghost, ghosts[ghostNum].ghostAx,
		   ghosts[ghostNum].ghostAy, 0, GHOSTSCALE, GHOSTSCALE,
		   0x40000002, aGhostImg[ghostIndex].transColor, 1) < 0)
    SPRModifySprite(ghosts[ghostNum].ghost, ghosts[ghostNum].ghostAx,
		    ghosts[ghostNum].ghostAy, 0, GHOSTSCALE, GHOSTSCALE,
		    0x40000002, aGhostImg[ghostIndex].transColor, 1);
  SPRSetSpriteSource(ghosts[ghostNum].ghost, ghostNum * SPRITESIZE, 0, SPRITESIZE, SPRITESIZE);
}

/*
  Initialize all sprites
*/

void InitSprites()
{
  int i;
  for(i = 0; i < numGhosts; i++) InitGhost(i);
  
  chase = 0;
  chomper.direction = 0;
  chomper.nchomperx = chomper.chomperx = chomper.startx;
  chomper.nchompery = chomper.chompery = chomper.starty;
  chomper.chomperAx = chomper.chomperx * TILESIZE + OFFX;
  chomper.chomperAy = chomper.chompery * TILESIZE + OFFY;
  chomper.incx = chomper.incy = chomper.mphase = 0;
  chomper.animDelay = ANIM_DELAY;
  if (SPRAddSprite(chomper.chomperman,
		   chomper.chomperx * TILESIZE + OFFX, chomper.chompery * TILESIZE + OFFY,
		   0, CHOMPERSCALE, CHOMPERSCALE, 0x40000002, kBlack, 1) < 0)
    SPRModifySprite(chomper.chomperman,
		    chomper.chomperx * TILESIZE + OFFX, chomper.chompery * TILESIZE + OFFY,
		    0, CHOMPERSCALE, CHOMPERSCALE, 0x40000002, kBlack, 1);
}

/*
  Wait for a joystick (or remote) button
*/

void WaitForJoystick(int but)
{
  int edge,joystick;
  int old_joystick=0;
  
  do {
    joystick = _Controller[1].buttons |_Controller[0].buttons;
    edge = (joystick ^ old_joystick) & joystick; 
    old_joystick = joystick;
  } while (!(edge & but));
}

/*
  Draw Score
*/

void DrawScore(char *frame, int x, int y)
{
  char temp[256], *s;
  
#if 1
  sprintf(temp,"%05d",score);
#else
  sprintf(temp,"%d",fps);
#endif
  s = temp;
  while(*s) {
    SPRBlitter(digitsImg.spr.img, digitsImg.spr.dmaflags, 1 + (int) (*s - '0') * 32, 1, 30, 30,
	       frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0x10000,
	       0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, digitsImg.transColor);
    s++;
    x += (24 * SCORESCALE) / 0x10000;
  }
  
  x += 24 * SCORESCALE / 0x10000;
  SPRBlitter(digitsImg.spr.img, digitsImg.spr.dmaflags, 1 + lives * 32, 1, 30, 30,
	     frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0x10000,
	     0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, digitsImg.transColor);

  x += 24 * SCORESCALE / 0x10000;
  SPRBlitter(digitsImg.spr.img, digitsImg.spr.dmaflags, 384, 0, 32, 32,
	     frame, DMAFLAGS, x, y, SCORESCALE, SCORESCALE, 0x10000,
	     0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, digitsImg.transColor);

  x += 24 * SCORESCALE / 0x10000;
  SPRBlitter(aPlayerImg[0].spr.img, aPlayerImg[0].spr.dmaflags, SPRITESIZE, 0,
	     SPRITESIZE, SPRITESIZE, frame, DMAFLAGS, x, y, LIVESSCALE, LIVESSCALE, 0x10000,
	     0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 0x40000002, kBlack);
}

void PauseGame (void){
  SPRAddSprite (PauseSprite, SCREEN_WIDTH/2, SCREEN_HEIGHT/2, 0, 0x9000, 0x7000, 0x40000002,
		kBlack, 1);
  SPRDraw (0,1);
  VideoSync();
  //	_VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[0]);
  PCMVoiceOff(chasevoice);
  WaitForJoystick(JOY_A);
  if (chase) chasevoice=PlayWAV((void *) ChaseWav, 1, &Pan, 0x40000000, 0x40000000);
  SPRRemoveSprite (PauseSprite);
}

//==================================================================================================
// showProgress
//
// Spins the progress indicator and draws it to the screen.
//
// Revision history
//   SDC 2002-02-10:  Initial revision
//==================================================================================================

void showProgress()
{
  VideoSync();
  progressRot += 0x1000;
  SPRSetSpriteRotation(sprProgress, progressRot);
  SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH-1, SCREEN_HEIGHT-1,kBlack);
  SPRDraw(0,0);
}

/*
  Main loop
*/

int main()
{
  int i, ret, next, pinc;
  VidDisplay display;
  VidChannel mainchannel;
  
  // Setup audio
  
  AUDIOInit();
  
  Pan.PCMPanLR = 0;
  Pan.PCMPanFB = 0;
  Pan.PCMPanUD = 0;
  
  // Get some local memory for DMA
  
  localScratch=_MemLocalScratch(&localSize);
  
  // Setup video
  
  for(i = 0; i < NUM_BUFFERS; i++)
    FrameBuffer[i]=_MemAlloc(SCREEN_WIDTH * SCREEN_HEIGHT * 4, 512, kMemSDRAM);
  MazeBuf = _MemAlloc(SCREEN_WIDTH * SCREEN_HEIGHT * 4, 512, kMemSDRAM);
  
  myFillRect(MazeBuf, DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
  
  display.dispwidth   = -1;
  display.dispheight  = -1;
  display.bordcolor   = 0x20100000;
  display.progressive = 0;
  for(i = 0; i < 6; i++) display.reserved[i] = 0;
  
  mainchannel.dmaflags    = DMAFLAGS;
  mainchannel.base        = FrameBuffer[0];
  mainchannel.dest_xoff   = -1;
  mainchannel.dest_yoff   = -1;
  mainchannel.dest_width  = 720;
  mainchannel.dest_height = 480;
  mainchannel.src_xoff    = 0;
  mainchannel.src_yoff    = 0;
  mainchannel.src_width   = SCREEN_WIDTH;
  mainchannel.src_height  = SCREEN_HEIGHT;
  mainchannel.clut_select = 0;
  mainchannel.alpha       = 0;
  mainchannel.vfilter     = VID_VFILTER_2TAP;
  mainchannel.hfilter     = VID_HFILTER_4TAP;
  for(i = 0; i < 5; i++) mainchannel.reserved[i] = 0;
  
  InitTerminalX(0,(int)FrameBuffer[0],SCREEN_WIDTH, SCREEN_HEIGHT, DMAFLAGS, kBlack);
  _VidConfig ( &display, &mainchannel, 0, 0L); 
  
  srand(_VidSync(-1));
  
  // Use MPEs 0,1 and 2 for rendering, 16 pixel slices
  
  SPRInit(0, 2, 16);
  
  // Set to the second framebuffer
  
  SPRSetDestScreen(FrameBuffer[1], DMAFLAGS,  0,  0,  SCREEN_WIDTH-1,  SCREEN_HEIGHT-1, kBlack);
  
  myFillRect(FrameBuffer[0], DMAFLAGS, 0, 0, 720, 480, kBlack);
  myFillRect(FrameBuffer[1], DMAFLAGS, 0, 0, 720, 480, kBlack);

  // Load image with progress indicator first
  miscImg.transColor = getFirstTGAPixel(miscImages);
  retVal = SPRInstallTGAImage(miscImages, 1, &miscImg.spr, miscImg.transColor);
  //  sprProgress = SPRCreateSprite(&miscImg.spr, PROGRESS_OFFSET_X, 0, 31, 31);
  //  SPRAddSprite(sprProgress, (SCREEN_WIDTH-SPRITESIZE)/2, (SCREEN_HEIGHT-SPRITESIZE)/2, 0,
  //	       1<<16, 1<<16, 0x40000002, miscImg.transColor, 1);
  //  progressRot = 0;

  // SDC 2002-02-10:  Random sprites
  // Find out how many of each type we have
  for (wallMax  = 0; aWallImg [wallMax ].tgaAddr; wallMax++ ) aWallImg[wallMax].init   = FALSE;
  for (pillMax  = 0; aPillImg [pillMax ].tgaAddr; pillMax++ ) aPillImg[pillMax].init   = FALSE;
  for (fruitMax = 0; aFruitImg[fruitMax].tgaAddr; fruitMax++) aFruitImg[fruitMax].init = FALSE;
  for (ghostMax = 0; aGhostImg[ghostMax].tgaAddr; ghostMax++) aGhostImg[ghostMax].init = FALSE;

  //  showProgress();

  // Install the TGA images

  aPlayerImg[0].transColor = getFirstTGAPixel(playerImages);
  retVal = SPRInstallTGAImage(playerImages,        1, &aPlayerImg[0].spr, 0x80407F00);

  //  showProgress();
  /*
  for (i = 0; i < wallMax; i++) {
    aWallImg[i].transColor = getFirstTGAPixel(aWallImg[i].tgaAddr);
    retVal = SPRInstallTGAImage(aWallImg[i].tgaAddr,  1, &aWallImg[i].spr, aWallImg[i].transColor);
    showProgress();
  }

  for (i = 0; i < pillMax; i++) {
    aPillImg[i].transColor = getFirstTGAPixel(aPillImg[i].tgaAddr);
    retVal = SPRInstallTGAImage(aPillImg[i].tgaAddr,  1, &aPillImg[i].spr, aPillImg[i].transColor);
    showProgress();
  }

  for (i = 0; i < fruitMax; i++) {
    aFruitImg[i].transColor = getFirstTGAPixel(aFruitImg[i].tgaAddr);
    retVal = SPRInstallTGAImage(aFruitImg[i].tgaAddr, 1, &aFruitImg[i].spr,aFruitImg[i].transColor);
    showProgress();
  }

  for (i = 0; i < ghostMax; i++) {
    aGhostImg[i].transColor = getFirstTGAPixel(aGhostImg[i].tgaAddr);
    retVal = SPRInstallTGAImage(aGhostImg[i].tgaAddr, 1, &aGhostImg[i].spr,aGhostImg[i].transColor);
    showProgress();
  }
  */

  retVal = SPRInstallTGAImage(StartScreen,         1, &StartScr,      0x0       );
  //  showProgress();
  retVal = SPRInstallTGAImage(GameOverScreen,      1, &GameOverScr,   0x0       );
  //  showProgress();
  retVal = SPRInstallTGAImage(GameCompleteScreen,  1, &CompleteScr,   0x0       );
  //  showProgress();
  retVal = SPRInstallTGAImage(LevelCompleteScreen, 1, &LevelScr,      0x0       );
  //  showProgress();
  digitsImg.transColor = getFirstTGAPixel(digitsImg.tgaAddr);
  retVal = SPRInstallTGAImage(Digits,              1, &digitsImg.spr, digitsImg.transColor);
  //  showProgress();
  retVal = SPRInstallTGAImage(PauseImage,          1, &PauseScr,      0x04020400);

  //  SPRRemoveSprite(sprProgress);
  //  SPRDeleteSprite(sprProgress);

  // Create sprites
  
  chomper.chomperman = SPRCreateSprite(&aPlayerImg[0].spr, SPRITESIZE, 0, SPRITESIZE, SPRITESIZE);
  PauseSprite = SPRCreateSprite(&PauseScr, 0, 0, 256, 128);

  // main game loop
  
  for(;;) {
      
    // Start screen

    myFillRect(FrameBuffer[0], DMAFLAGS, 0, 0, 720, 480, kBlack);
    _VidChangeBase(VID_CHANNEL_MAIN,DMAFLAGS,FrameBuffer[0]);
      
    SPRBlitter(StartScr.img, StartScr.dmaflags, 0, 0, 600, 400, FrameBuffer[0], DMAFLAGS,
	       SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
	       0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480,
	       0, 0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
      
    WaitForJoystick(JOY_START);
    while (ButtonStart(_Controller[1]));

    debug = ((_Controller[1].buttons | _Controller[0].buttons) & JOY_A) ? 1 : 0;

    score = 0;
    lives = 2;
    level = 0;
    extra = 10000;  // next extra life at 10000 points
    ret = PlayWAV((void *) GameBeginWav, -1, &Pan, 0x40000000, 0x40000000);
      
    for(i = 0; i < MAX_GHOSTS; i++) ghosts[i].ghost = NULL;

    // Level loop
      
    do {
      collected = 0;
      delay     = 0;
      pDividerCnt = pDivider = 1;

      if (debug) {
	gDividerCnt = gDivider = levelsDebug[level].gDivider;
	memcpy(maze, levelsDebug[level].maze, MAZE_W * MAZE_H);
      } else {
	gDividerCnt = gDivider = levels[level].gDivider;
	memcpy(maze, levels[level].maze, MAZE_W * MAZE_H);
      }

      // Clear level before drawing (not necessary?)
      myFillRect(MazeBuf, DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);

      // SDC 2002-02-10:  Choose sprites at random for each level
      wallIndex  = rand() % wallMax;
      pillIndex  = rand() % pillMax;
      fruitIndex = rand() % fruitMax;
      ghostIndex = rand() % ghostMax;

      if (!aWallImg[wallIndex].init) {
	aWallImg[wallIndex].transColor = getFirstTGAPixel(aWallImg[wallIndex].tgaAddr);
	retVal = SPRInstallTGAImage(aWallImg[wallIndex].tgaAddr, 1, &aWallImg[wallIndex].spr,
				    aWallImg[wallIndex].transColor);
	aWallImg[wallIndex].init = TRUE;
      }

      if (!aPillImg[pillIndex].init) {
	aPillImg[pillIndex].transColor = getFirstTGAPixel(aPillImg[pillIndex].tgaAddr);
	retVal = SPRInstallTGAImage(aPillImg[pillIndex].tgaAddr, 1,&aPillImg[pillIndex].spr,
				    aPillImg[pillIndex].transColor);
	aPillImg[pillIndex].init = TRUE;
      }

      if (!aFruitImg[fruitIndex].init) {
	aFruitImg[fruitIndex].transColor = getFirstTGAPixel(aFruitImg[fruitIndex].tgaAddr);
	retVal = SPRInstallTGAImage(aFruitImg[fruitIndex].tgaAddr, 1, &aFruitImg[fruitIndex].spr,
				    aFruitImg[fruitIndex].transColor);
	aFruitImg[fruitIndex].init = TRUE;
      }

      if (!aGhostImg[ghostIndex].init) {
	aGhostImg[ghostIndex].transColor = getFirstTGAPixel(aGhostImg[ghostIndex].tgaAddr);
	retVal = SPRInstallTGAImage(aGhostImg[ghostIndex].tgaAddr, 1, &aGhostImg[ghostIndex].spr,
				    aGhostImg[ghostIndex].transColor);
	aGhostImg[ghostIndex].init = TRUE;
      }

      // Need to delete and recreate ghosts since sprite changes every level
      for(i = 0; i < MAX_GHOSTS; i++) {
	if (ghosts[i].ghost) {
	  SPRRemoveSprite(ghosts[i].ghost);
	  SPRDeleteSprite(ghosts[i].ghost);
	}
	ghosts[i].ghost = SPRCreateSprite(&aGhostImg[ghostIndex].spr, i * SPRITESIZE, 0,
					  SPRITESIZE, SPRITESIZE);
      }

      // Draw the maze given the current randomly selected sprites
      CreateMaze(&aWallImg[wallIndex].spr, &aPillImg[pillIndex].spr,
		 &aFruitImg[fruitIndex].spr, MazeBuf, maze, MAZE_W, MAZE_H);

      InitSprites();

      for(i = 0; i < NUM_BUFFERS; i++) {
	myFillRect(FrameBuffer[i], DMAFLAGS, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT, kBlack);
	CopyMaze(FrameBuffer[i], MazeBuf, SCREEN_WIDTH * SCREEN_HEIGHT * 4 / 256);
      }

      next = chomper.aphase = 0;
      pinc = 1;

      // Inner game play loop

      for(;;) {

	// My PAUSE feature (by using start button) 1/14/2002
	if (ButtonStart(_Controller[1])) PauseGame();
	      
	if (score > extra) {
	  PlayWAV((void *) ExtraChomperWav, -1, &Pan, 0x40000000, 0x40000000);
	  lives++;
	  extra += 12000;
	}

	if (delay) _TimeToSleep(delay);

	// Check if level done
	if (collected == pills)	{

	  // Level is done

	  // Quiet things down
	  PCMVoiceOff(chasevoice);

	  // Advance to next level unless we're in debug mode
	  if (!debug) level++;

	  // Remove spinning score sprite
	  if (saphase) {  SPRRemoveSprite(bonusSprite);  saphase = 0;  }

	  // Update final score
	  DrawScore(FrameBuffer[drawBuffer], SCOREX, SCOREY);

	  // Check if done with last level
	  if (level == NUMLEVELS) {
	    SPRBlitter(CompleteScr.img, CompleteScr.dmaflags, 0, 0, 256, 128,
		       FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
		       0x18000 * SCREEN_WIDTH / 720, 0x18000 * SCREEN_HEIGHT / 480, 0,
		       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
		      
	  } else {
	    SPRBlitter(LevelScr.img, LevelScr.dmaflags, 0, 0, 256, 128,
		       FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
		       0x18000 * SCREEN_WIDTH / 720, 0x18000 * SCREEN_HEIGHT / 480, 0,
		       0, 0, SCREEN_WIDTH - 1, SCREEN_HEIGHT - 1, 0x40000002, 0);
	  }
	  VideoSync();
	  WaitForJoystick(JOY_A);
	  break;
	}

	// Are we in "power pill" mode?
	if (chase) {
	  // Fade ghost colors in & out
	  for(i = 0; i < numGhosts; i++)
	    if (ghosts[i].active)
	      SPRSetSpriteType(ghosts[i].ghost,0x2+(chase%32)*0x2000000+0x2000000);
	    else
	      SPRSetSpriteType(ghosts[i].ghost,0x40000002);
	  chase--;
	  if (!chase) {
	    // Exiting "power pill" mode - return sprites to normal color
	    for(i = 0; i < numGhosts; i++) SPRSetSpriteType(ghosts[i].ghost, 0x40000002);
	    PCMVoiceOff(chasevoice);
	  }
	}

	if (gameover) {
	  // Game over
		  
	  // This waits until the "death" sound is done
	  if (PCMGetUsedVoices()) {
	    // Rotate chomperman while "dying" sound is playing
	    SPRSetSpriteRotation(chomper.chomperman, gameover * 0x800);
	    SPRSetSpriteScale(chomper.chomperman, CHOMPERSCALE - gameover * 0x100 / RATIO,
			      CHOMPERSCALE - gameover * 0x100 / RATIO);
	    gameover++;

	  } else {
	    // Sound is done, reinit player
	    lives--;
	    gameover = 0;
	    InitSprites();
	    if (!lives) {
	      DrawScore(FrameBuffer[drawBuffer], 80, 420);
	      SPRBlitter(GameOverScr.img, GameOverScr.dmaflags, 0, 0, 256, 128,
			 FrameBuffer[drawBuffer], DMAFLAGS, SCREEN_WIDTH/2, SCREEN_HEIGHT/2,
			 0x10000 * SCREEN_WIDTH / 720, 0x10000 * SCREEN_HEIGHT / 480, 0,
			 0, 0, SCREEN_WIDTH - 1,SCREEN_HEIGHT - 1, 0x40000002, 0);
			  
	      VideoSync();
	      WaitForJoystick(JOY_A);

	      // Remove all ghosts - prevents leftovers on restart
	      for(i = 0; i < numGhosts; i++) SPRRemoveSprite(ghosts[i].ghost);
			  
	      level = NUMLEVELS; // force restart
	      break;
	    }
	  }
	} else {
	  //
	  // Normal loop
	  //

	  // Move Chomperman
	  ControlChomperman();

	  // Move ghosts based on their current speed
	  gDividerCnt--;
	  if (!gDividerCnt) {
	    for (i = 0; i < numGhosts; i++) ControlGhost(i);
	    gDividerCnt = gDivider;
	  }
	  DoScoreAnimation();

	  // SDC 2002-02-10:  Parameterized number of animation frames and increased to 4
	  // SDC 2002-02-10:  Sped up animation to 1 per frame
	  // SDC 2002-02-16:  Slowed down to every other frame
	  if (chomper.direction) {
	    if (!chomper.animDelay) {
	      SPRSetSpriteSource(chomper.chomperman, chomper.aphase * SPRITESIZE, 0,
				 SPRITESIZE, SPRITESIZE);
	      chomper.aphase += pinc;
	      if ((chomper.aphase == 0) || (chomper.aphase == (CHOMPER_ANIM_FRAMES-1))) pinc = -pinc;
	      chomper.animDelay = ANIM_DELAY;
	    } else {
	      chomper.animDelay--;
	    }
	  }
	}
	VideoSync();
	      
	// Copy maze into frame buffer
	CopyMaze(FrameBuffer[drawBuffer], MazeBuf, SCREEN_WIDTH * SCREEN_HEIGHT * 4 / 256);
	      
	// Print score   
	DrawScore(FrameBuffer[drawBuffer], SCOREX, SCOREY);

	// Draw chase progress bar
	if (chase)
	  myFillRect(FrameBuffer[drawBuffer], DMAFLAGS, CHASEBARX, CHASEBARY, chase / RATIO,
		     TILESIZE, kGreen);
	      
	// Process display list
	SPRSetDestScreen(FrameBuffer[drawBuffer], DMAFLAGS, 0, 0, SCREEN_WIDTH - 1,
			 SCREEN_HEIGHT - 1, kBlack);
	SPRDraw(0,0);
	      
	SPRWait();
      }
    } while (level < NUMLEVELS);
  }
}
